import { Client } from "@notionhq/client";
import { mkdirSync, createWriteStream, writeFileSync } from 'fs';
import { join } from 'path';
import { get } from 'https';

import * as dotenv from 'dotenv';
dotenv.config({ path: join(process.cwd(), '.env') });

export const workspacePath = join(process.cwd(), process.env.WORKSPACE_PATH);

const cleanTitle = (title) => {
  return title.match(/[A-Z0-9]+/ig).join('-');
}

const notion = new Client({
  auth: process.env.NOTION_TOKEN,
});

const getDatabase = async (databaseId) => {
  return await notion.databases.retrieve({ database_id: databaseId });
}

const getDatabaseItems = async (databaseId) => {
  const response = await notion.databases.query({
    database_id: databaseId,
  });
  return response.results.map((page) => ({ id: page.id }));
}

const getFile = async (file, id) => {
  const { type } = file;
  const filePath = join(process.cwd(), `public/files/${id}`);
  const fileWriteStream = createWriteStream(filePath);
  get(file[type].url, (res) => {
    res.pipe(fileWriteStream);
    fileWriteStream.on("finish", () => {
      fileWriteStream.close();
    });
  })
  return `/files/${id}`;
}

const getPage = async (pageId) => {
  const page = await notion.pages.retrieve({ page_id: pageId });
  for (const entry of Object.entries(page.properties)) {
    const [property, { id }] = entry;
    const props = await notion.pages.properties.retrieve({
      page_id: pageId,
      property_id: id
    });
    if (id === "title") {
      page.properties.title = props.results[0].title;
    } else {
      page.properties[property] = props.results;
    }
  }
  if (page.cover) {
    page.cover.filePath = await getFile(page.cover, `cover-${pageId}`);
  }
  if (page.icon) {
    page.icon.filePath = await getFile(page.icon, `icon-${pageId}`);
  }
  return page;
};

const resolveParent = (objWithParent, title, workspace) => {
  const { parent: { type } } = objWithParent;
  const parentId = objWithParent.parent[type];
  let nextPath;
  if (type == "database_id") {
    const parentDb = workspace.databases[parentId].database;
    const parentNextPath = parentDb.is_inline 
      ? workspace.pages[parentDb.parent.page_id].nextPath
      : parentDb.nextPath;
    nextPath = [...parentNextPath, title];
  } else if (type == "page_id") {
    const parentNextPath = workspace.pages[parentId].nextPath;
    nextPath = [...parentNextPath, title]
  } else {
    nextPath = [];
  }
  return nextPath;
}

async function crawlBlocks(blockId, workspace) {
  const blocks = [];
  let cursor;
  while (true) {
    const { results, next_cursor } = await notion.blocks.children.list({
      start_cursor: cursor,
      block_id: blockId,
    });
    blocks.push(...results);
    if (!next_cursor) {
      break;
    }
    cursor = next_cursor;
  }
  for (const block of blocks) {
    const { id, type, has_children } = block;
    if (type === "child_page") {
      await crawlPage(id, workspace);
    } else if (type === "child_database") {
      await crawlDatabase(id, workspace);
    } else if (type === "image"
              || type === "file"
              || type === "pdf") {
      const file = block[type];
      file.filePath = await getFile(file, id);
    } else if (has_children) {
      block.children = await crawlBlocks(id, workspace);
    }
  }
  return blocks;
}

async function crawlDatabase(databaseId, workspace) {
  if (databaseId in workspace.databases)
    return;
  const database = await getDatabase(databaseId);
  const pages = await getDatabaseItems(databaseId);
  const title = cleanTitle(database.title[0].plain_text);
  const nextPath = database.is_inline ? [title] : resolveParent(database, title, workspace);
  const nextPathStr = nextPath.join('/');
  workspace.databases[databaseId] = {
    nextPath,
    database,
    pages,
  }
  workspace.pathDatabaseId[nextPathStr] = databaseId;
  for (const page of pages) {
    await crawlPage(page.id, workspace)
    for (const entry of Object.entries(database.properties)) {
      const finishedPage = workspace.pages[page.id].page;
      const [property, { id }] = entry;
      if (typeof finishedPage.properties[property] !== 'undefined')
        continue;
      const props = await notion.pages.properties.retrieve({
        page_id: page.id,
        property_id: id
      });
      finishedPage.properties[property] = props;
    }
  }
}


async function crawlPage(pageId, workspace) {
  if (pageId in workspace.pages)
    return;
  const page = await getPage(pageId);
  const title = cleanTitle(page.properties.title.plain_text);
  const nextPath = resolveParent(page, title, workspace);
  const nextPathStr = nextPath.join('/');
  workspace.pages[pageId] = {
    nextPath,
    nextPathStr,
    page
  }
  workspace.pathPageId[nextPathStr] = pageId;
  console.log(`Syncing /${nextPathStr}`);
  const blocks = await crawlBlocks(pageId, workspace);
  workspace.pages[pageId].blocks = blocks;
}
    

export const init = async () => {
  mkdirSync('./public/files', { recursive: true });
  console.log(`Sycning notion workspace to ${workspacePath}`);
  const rootId = process.env.ROOT_PAGE_ID;
  const workspace = { 
    pages: {}, 
    databases: {}, 
    pathPageId: {}, 
    pathDatabaseId: {} 
  };
  // rootId must point to a page. 
  await crawlPage(rootId, workspace);
  writeFileSync(workspacePath, JSON.stringify(workspace));
  console.log("COMPLETE");
}
